class LocalFileNotFound(Exception):
    """
    This file could not be found locally, when we expected it to be.
    """
    pass


class LocalCacheDirNotFound(Exception):
    """
    We couldn't find a local cache dir to use
    """
    pass


class FileNotFound(Exception):
    """
    This file could not be found locally or remotely
    """
    pass


class ConnectionError(Exception):
    """
    Error connecting to SFTP server
    """
    pass


class TransferError(Exception):
    """
    Connected, but some other problem transferring file
    """
    pass


class RemoteDirNotCreated(Exception):
    """
    Remote dir could not be created
    """
    pass



from setuptools import setup

setup(
    name='sftpsyncer',
    version='0.1.1',
    author='Chris Horsley',
    author_email='chris.horsley@csirtfoundry.com',
    packages=['sftpsyncer', 'sftpsyncer.test'],
    scripts=[],
    url='http://bitbucket.org/dsms/sftpsyncher/',
    license='LICENSE.txt',
    description='Synchronise files between a local cache and an SFTP server',
    long_description=open('README').read(),
    install_requires=[
        "paramiko"
    ],
)



# -*- coding: utf-8 -*-

import unittest
from subprocess import Popen, check_output
import time
import sys
import os
from contextlib import contextmanager
import uuid
import paramiko
import logging

from sftpsyncer.sftpagent import SyncAgent
import sftpsyncer.exceptions

logger = logging.getLogger(name=__name__)

try:
    import sftpserver  # noqa
except:
    sys.exit("\nYou're missing one or more test requirements. Please install "
             "via:\n\npip install -r test_requirements.txt")

KEY = "/tmp/sftpsyncer_rsakey_%s" % uuid.uuid4()
KEYPASS = ""
create_key = check_output(
    ["ssh-keygen", "-b", "2048", "-t", "rsa", "-f", KEY, "-N", KEYPASS])
PORT = 3373


@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(newdir)
    try:
        yield
    finally:
        os.chdir(prevdir)


def create_tmp_dir(dir_type):
    dirname = generate_tmp_dir_name(dir_type)
    os.mkdir(dirname)
    return dirname


def generate_tmp_dir_name(dir_type):
    dirname = "/tmp/sftpsynctest_%s_%s" % (dir_type, uuid.uuid4())
    return dirname


class TestFileOps(unittest.TestCase):
    def setUp(self):
        self.sftpproc = None

        sftpbin = check_output(["which", "sftpserver"]).replace("\n", "")
        self.sftp_file_root = generate_tmp_dir_name("sftproot")
        os.mkdir(self.sftp_file_root)
        with cd(self.sftp_file_root):
            self.sftpproc = Popen([sftpbin, "--host", "localhost", "-k", KEY,
                                   "-p", str(PORT), "-l", "WARNING"])
        time.sleep(0.1)

    def tearDown(self):
        self.sftpproc.terminate()

    def _localised_remote_path(self, path):
        return os.path.join(self.sftp_file_root, path.lstrip("/"))

    def _localised_remote_file_path(self, path, file):
        return os.path.join(self._localised_remote_path(path), file)

    def test_sftpserver_working(self):
        # pkey = paramiko.RSAKey.from_private_key_file(KEY, KEYPASS)
        transport = paramiko.Transport(('localhost', PORT))
        transport.connect(username='admin', password="")
        sftp = paramiko.SFTPClient.from_transport(transport)
        dirs = sftp.listdir('.')
        self.assertEqual(dirs, [])

    def test_connection_failure(self):
        remote_root = "/repo"
        local_root = create_tmp_dir("localroot")

        try:
            sa = SyncAgent(hostname="something.localhost", host_port=60000,
                           username="theuser", key_path=KEY,
                           local_root=local_root, remote_root=remote_root)
            sa.connect()
        except sftpsyncer.exceptions.ConnectionError:
            assert(True)
        else:
            assert(False)

    def test_no_such_local_dir(self):
        remote_root = "/repo"
        local_root = "No such dir here"

        try:
            SyncAgent(hostname="localhost", host_port=PORT,
                      username="theuser", key_path=KEY, local_root=local_root,
                      remote_root=remote_root)
        except sftpsyncer.exceptions.LocalCacheDirNotFound:
            assert(True)
        else:
            assert(False)

    def test_mkdir_p_basic(self):
        local_root = create_tmp_dir("localroot")
        remote_root = "/my/test/dir/"

        sa = SyncAgent(hostname="localhost", host_port=PORT,
                       username="theuser", key_path=KEY, local_root=local_root,
                       remote_root=remote_root)

        sa.mkdir_p(remote_root)
        remote_absolute_path = self._localised_remote_path(remote_root)
        logger.debug(self.sftp_file_root)
        logger.debug("Remote target path: %s" % remote_absolute_path)
        self.assertTrue(os.path.isdir(remote_absolute_path))

    def test_mkdir_p_empty(self):
        local_root = create_tmp_dir("localroot")
        remote_root = ""

        sa = SyncAgent(hostname="localhost", host_port=PORT,
                       username="theuser", key_path=KEY, local_root=local_root,
                       remote_root=remote_root)

        try:
            sa.mkdir_p(remote_root)
        except sftpsyncer.exceptions.RemoteDirNotCreated:
            assert(True)
        else:
            assert(False)

    def test_mkdir_p_invalid_chars(self):
        pass

    def test_local_no_file_sync(self):
        remote_root = "/repo"
        local_root = create_tmp_dir("localroot")

        sa = SyncAgent(hostname="localhost", host_port=PORT,
                       username="theuser", key_path=KEY, local_root=local_root,
                       remote_root=remote_root)
        sa.push()

    def test_local_single_sync_unicode_file(self):
        file_name = "香港.txt"
        remote_root = "/repo"
        local_root = create_tmp_dir("localroot")

        with open(os.path.join(local_root, file_name), "wb") as f:
            f.write("test data")

        sa = SyncAgent(hostname="localhost", host_port=PORT,
                       username="theuser", key_path=KEY, local_root=local_root,
                       remote_root=remote_root)
        sa.upload_queue.add(file_name)
        sa.push()

        remote_file_path = self._localised_remote_file_path(
            remote_root, file_name)
        remote_fh = open(remote_file_path, "r")
        self.assertEqual(remote_fh.read(), "test data")

    def test_local_single_unicode_dir(self):
        file_name = "myfile.txt"
        remote_root = "/日本"
        local_root = create_tmp_dir("localroot")

        with open(os.path.join(local_root, file_name), "wb") as f:
            f.write("test data")

        sa = SyncAgent(hostname="localhost", host_port=PORT,
                       username="theuser", key_path=KEY, local_root=local_root,
                       remote_root=remote_root)
        sa.upload_queue.add(file_name)
        sa.push()

        remote_file_path = self._localised_remote_file_path(
            remote_root, file_name)
        remote_fh = open(remote_file_path, "r")
        self.assertEqual(remote_fh.read(), "test data")

    def test_local_single_sync_root(self):
        file_name = "myfile.txt"
        remote_root = "/"
        local_root = create_tmp_dir("localroot")

        with open(os.path.join(local_root, file_name), "wb") as f:
            f.write("test data")

        sa = SyncAgent(hostname="localhost", host_port=PORT,
                       username="theuser", key_path=KEY, local_root=local_root,
                       remote_root=remote_root)
        sa.upload_queue.add(file_name)
        sa.push()

        remote_file_path = self._localised_remote_file_path(
            remote_root, file_name)
        remote_fh = open(remote_file_path, "r")
        self.assertEqual(remote_fh.read(), "test data")

    def test_local_multi_sync(self):
        file_names = ["myfile.txt", "anotherfile.txt"]
        remote_root = "/repo"
        local_root = create_tmp_dir("localroot")

        for file_name in file_names:
            with open(os.path.join(local_root, file_name), "w") as f:
                f.write("test data")

        sa = SyncAgent(hostname="localhost", host_port=PORT,
                       username="theuser", key_path=KEY, local_root=local_root,
                       remote_root=remote_root)
        for file_name in file_names:
            sa.upload_queue.add(file_name)
        sa.push()

        for file_name in file_names:
            remote_file_path = self._localised_remote_file_path(
                remote_root, file_name)
            remote_fh = open(remote_file_path, "r")
            self.assertEqual(remote_fh.read(), "test data")

    def test_local_single_sync(self):
        file_name = "myfile.txt"
        remote_root = "/repo"
        local_root = create_tmp_dir("localroot")

        with open(os.path.join(local_root, file_name), "wb") as f:
            f.write("test data")

        sa = SyncAgent(hostname="localhost", host_port=PORT,
                       username="theuser", key_path=KEY, local_root=local_root,
                       remote_root=remote_root)
        sa.upload_queue.add(file_name)
        sa.push()

        remote_file_path = self._localised_remote_file_path(
            remote_root, file_name)
        remote_fh = open(remote_file_path, "r")
        self.assertEqual(remote_fh.read(), "test data")

    def test_local_dir_push(self):
        remote_root = "/deep"
        local_root = create_tmp_dir("localroot")

        sa = SyncAgent(hostname="localhost", host_port=PORT,
                       username="theuser", key_path=KEY, local_root=local_root,
                       remote_root=remote_root)

        dir_nest = ""
        for dir_ in ["a", "b", "c"]:
            dir_nest = dir_nest + dir_ + "/"
            local_dir_path = os.path.join(local_root, dir_nest)
            os.mkdir(local_dir_path)
            for file_ in ["1", "2", "3"]:
                local_file_path = os.path.join(local_dir_path, file_)
                logger.debug(local_file_path)
                with open(local_file_path, "w") as f:
                    f.write("test data")
        sa.upload_queue.add_dir(".")
        sa.push()

        dir_nest = ""
        for dir_ in ["a", "b", "c"]:
            dir_nest = dir_nest + dir_ + "/"

            for file_ in ["1", "2", "3"]:
                remote_abs_file_path = self._localised_remote_file_path(
                    os.path.join(remote_root, dir_nest), file_)
                logger.debug(remote_abs_file_path)
                with open(remote_abs_file_path, "r") as f:
                    self.assertEqual(f.read(), "test data")

    def test_local_dir_pull(self):
        remote_root = "/deep"
        local_root = create_tmp_dir("localroot")

        sa = SyncAgent(hostname="localhost", host_port=PORT,
                       username="theuser", key_path=KEY, local_root=local_root,
                       remote_root=remote_root)

        # create files in remote SFTP server
        dir_nest = ""
        for dir_ in ["a", "b", "c"]:
            dir_nest = dir_nest + dir_ + "/"
            remote_abs_file_path = self._localised_remote_path(
                os.path.join(remote_root, dir_nest))
            logger.debug("makedirs %s" % remote_abs_file_path)
            os.makedirs(remote_abs_file_path)
            for file_ in ["1", "2", "3"]:
                remote_abs_file_path = self._localised_remote_file_path(
                    os.path.join(remote_root, dir_nest), file_)
                logger.debug(remote_abs_file_path)
                with open(remote_abs_file_path, "w") as f:
                    f.write("test data")
        sa.pull_dir("*")

        # verify local cache files
        dir_nest = ""
        for dir_ in ["a", "b", "c"]:
            dir_nest = dir_nest + dir_ + "/"
            local_dir_path = os.path.join(local_root, dir_nest)
            for file_ in ["1", "2", "3"]:
                local_file_path = os.path.join(local_dir_path, file_)
                logger.debug("Testing local file %s" % local_file_path)
                with open(local_file_path, "r") as f:
                    self.assertEqual(f.read(), "test data")

    def test_get_remote_file(self):
        file_name = "myfile.txt"
        remote_root = "/repo/other/dir"
        local_root = create_tmp_dir("localroot")

        os.makedirs(self._localised_remote_path(remote_root))
        # simulate a file already there
        remote_abs_path = self._localised_remote_file_path(
            remote_root, file_name)
        logger.debug("Remote localised path: %s" % remote_abs_path)
        with open(remote_abs_path, "w") as f:
            f.write("test data")

        sa = SyncAgent(hostname="localhost", host_port=PORT,
                       username="theuser", key_path=KEY, local_root=local_root,
                       remote_root=remote_root)

        self.assertFalse(os.path.exists(os.path.join(local_root, file_name)))
        data = sa.open(file_name)
        self.assertEqual(data.read(), "test data")
        self.assertTrue(os.path.exists(os.path.join(local_root, file_name)))

    def test_get_remote_dir(self):
        pass

    def test_file_missing_everywhere(self):
        file_name = "myfile.txt"
        remote_root = "/repo"
        local_root = create_tmp_dir("localroot")

        sa = SyncAgent(hostname="localhost", host_port=PORT,
                       username="theuser", key_path=KEY, local_root=local_root,
                       remote_root=remote_root)

        self.assertFalse(os.path.exists(os.path.join(local_root, file_name)))
        try:
            sa.open(file_name)
        except sftpsyncer.exceptions.FileNotFound:
            assert(True)
        else:
            assert(False)

    def test_ignore_remote_if_cached(self):
        """
        If we have the file locally, we don't check remotely. Note that the
        file is not automatically uploaded - this must be done manually.
        """
        file_name = "myfile.txt"
        remote_root = "/repo"
        local_root = create_tmp_dir("localroot")

        with open(os.path.join(local_root, file_name), "wb") as f:
            f.write("test data")

        sa = SyncAgent(hostname="localhost", host_port=PORT,
                       username="theuser", key_path=KEY, local_root=local_root,
                       remote_root=remote_root)
        sa.upload_queue.add(file_name)
        sa.push()

        remote_fh = sa.open(file_name)
        self.assertEqual(remote_fh.read(), "test data")

    def test_race_condition_between_queue_and_transfer(self):
        file_name = "myfile.txt"
        remote_root = "repo"
        local_root = create_tmp_dir("localroot")
        local_file_path = os.path.join(local_root, file_name)

        with open(local_file_path, "w") as f:
            f.write("test data")

        sa = SyncAgent(hostname="localhost", host_port=PORT,
                       username="theuser", key_path=KEY, local_root=local_root,
                       remote_root=remote_root)
        sa.upload_queue.add(file_name)
        # file is remove before committing
        os.unlink(local_file_path)
        try:
            sa.push()
        except:
            assert(True)
        else:
            assert(False)



from functools import wraps
import logging
import paramiko
from stat import S_ISDIR
import os.path
from .exceptions import (LocalFileNotFound, FileNotFound, ConnectionError,
                         LocalCacheDirNotFound, TransferError,
                         RemoteDirNotCreated)

logger = logging.getLogger(name=__name__)

ROOT_ALIASES = ["*", "/", "."]


class SyncAgentLocalQueue(object):
    def __init__(self, local_root):
        self.queued_files = []

        self.local_root = local_root

    def add(self, rel_file_path):
        file_path = os.path.join(self.local_root, rel_file_path)

        if not os.path.exists(file_path):
            raise LocalFileNotFound(
                "Could not add file: %s not found" % file_path)

        self.queued_files.append(rel_file_path)

    def _add_wrapper(self, _, abs_dir_name, names):
        rel_dir_path = abs_dir_name.replace(self.local_root + "/", "")
        for name in names:
            abs_file_path = os.path.join(abs_dir_name, name)
            if os.path.isfile(abs_file_path):
                rel_file_path = os.path.join(rel_dir_path, name)
                logger.debug("Adding to package: %s" % rel_file_path)
                self.add(rel_file_path)

    def add_dir(self, rel_dir_path):
        dir_path = None
        if rel_dir_path in ROOT_ALIASES:
            # add everything under root
            dir_path = self.local_root
        else:
            dir_path = os.path.join(self.local_root, rel_dir_path)
        os.path.walk(dir_path, self._add_wrapper, None)

    def clear_queue(self):
        self.queued_files = []

    def __iter__(self):
        return self.queued_files.__iter__()


class SyncAgent(object):
    def __init__(self, hostname=None, host_port=22, username=None,
                 key_path=None, local_root=None, remote_root=None):
        self.hostname = hostname
        self.host_port = host_port
        self.username = username
        self.key_path = key_path
        self.local_root = local_root
        self.remote_root = remote_root

        if not os.path.isdir(self.local_root):
            raise LocalCacheDirNotFound(
                "Couldn't find local_root at %s: make sure this is "
                "created" % local_root)

        self.upload_queue = SyncAgentLocalQueue(local_root)
        self.conn = None

    def _rel_path_to_abs_local(self, path):
        path = path.lstrip("/")
        return os.path.join(self.local_root, path)

    def _rel_path_to_abs_remote(self, path):
        path = path.lstrip("/")
        return os.path.join(self.remote_root, path)

    def ensure_connected(func):
        @wraps(func)
        def connect_wrapper(self, *args, **kwargs):
            self.connect()
            return func(self, *args, **kwargs)
        return connect_wrapper

    def connect(self):
        if not self.conn:
            try:
                pkey = paramiko.RSAKey.from_private_key_file(self.key_path)
                transport = paramiko.Transport((self.hostname, self.host_port))
                transport.connect(username=self.username, pkey=pkey)
                transport.use_compression(compress=True)
                self.conn = paramiko.SFTPClient.from_transport(transport)
            except Exception as e:
                raise ConnectionError(
                    "Couldn't connect to SFTP server: %s" % e)

    @ensure_connected
    def push(self):
        for file in self.upload_queue:
            local_file_path = os.path.join(self.local_root, file)
            remote_temp_file_path = os.path.join(self.remote_root,
                                                 "%s.part" % file)
            remote_file_path = os.path.join(self.remote_root, file)
            try:
                assert(os.path.exists(local_file_path))
                remote_dir_path, _ = os.path.split(remote_file_path)
                self.mkdir_p(remote_dir_path)
                self.conn.put(local_file_path, remote_temp_file_path,
                              confirm=True)
                self.conn.rename(remote_temp_file_path, remote_file_path)
            except AssertionError as e:
                raise LocalFileNotFound("Error during commit: %s missing"
                                        % local_file_path)
            except Exception as e:
                raise TransferError("Couldn't transfer %s: %s" % (
                    local_file_path, e))

    @ensure_connected
    def pull_dir(self, rel_path):
        if rel_path in ROOT_ALIASES:
            # add everything under root
            remote_abs_dir_path = self.remote_root
            rel_path = ""
        else:
            remote_abs_dir_path = self._rel_path_to_abs_remote(rel_path)

        files = self.conn.listdir(remote_abs_dir_path)

        for file in files:
            remote_abs_file_path = os.path.join(remote_abs_dir_path, file)
            remote_rel_file_path = os.path.join(rel_path, file)
            logger.debug("local_file_path %s, %s, %s" %
                         (self.local_root, rel_path, file))
            local_file_path = os.path.join(
                self._rel_path_to_abs_local(rel_path), file)
            if self.is_remote_dir(remote_abs_file_path):
                logger.debug("mkdir %s and pull %s" %
                             (local_file_path, remote_abs_file_path))
                os.makedirs(local_file_path)
                self.pull_dir(remote_rel_file_path)
            else:
                logger.debug("Getting %s to %s" %
                             (remote_abs_file_path, local_file_path))
                self.conn.get(remote_abs_file_path, local_file_path)

    @ensure_connected
    def is_remote_dir(self, path):
        try:
            logger.debug("Is %s a dir?" % path)
            return S_ISDIR(self.conn.stat(path).st_mode)
        except IOError:
            logger.debug("Not a dir")
            return False

    @ensure_connected
    def mkdir_p(self, remote_path):
        logger.debug("mkdir_p:remote_path %s" % remote_path)
        if remote_path == '/':
            # absolute path so change directory to root
            self.conn.chdir('/')
            return
        if not remote_path:
            # top-level relative directory must exist
            raise RemoteDirNotCreated("You must specify a path")
        _, basename = os.path.split(remote_path)
        self.mkdir_p(os.path.dirname(remote_path))
        try:
            self.conn.chdir(basename)
        except IOError:
            logger.debug("mkdir_p:mkdir %s" % remote_path)
            self.conn.mkdir(basename)
            self.conn.chdir(basename)

    @ensure_connected
    def open(self, path, mode="r"):
        """
        Given a relative file path, check the local cache. If present, return a
        filehandle. If not in the local cache, check the remote SFTP server. If
        present remotely, get and place in the local cache. If missing
        remotely, raise a FileNotFound exception.
        """

        local_file_path = self._rel_path_to_abs_local(path)
        logger.debug("Checking local_file_path: %s" % local_file_path)
        if not os.path.isfile(local_file_path):
            remote_file_path = self._rel_path_to_abs_remote(path)
            logger.debug("Local file missing get remote: %s"
                         % remote_file_path)
            try:
                self.conn.get(remote_file_path, local_file_path)
            except IOError as e:
                raise FileNotFound("%s not found locally or remotely: %s" %
                                   (path, e))
        try:
            return open(local_file_path, mode)
        except IOError:
            raise
